package fr.toulouse.jug.languageserver;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Collectors;

import org.eclipse.lsp4j.CodeActionParams;
import org.eclipse.lsp4j.CodeLens;
import org.eclipse.lsp4j.CodeLensParams;
import org.eclipse.lsp4j.Command;
import org.eclipse.lsp4j.CompletionItem;
import org.eclipse.lsp4j.CompletionList;
import org.eclipse.lsp4j.Diagnostic;
import org.eclipse.lsp4j.DiagnosticSeverity;
import org.eclipse.lsp4j.DidChangeTextDocumentParams;
import org.eclipse.lsp4j.DidCloseTextDocumentParams;
import org.eclipse.lsp4j.DidOpenTextDocumentParams;
import org.eclipse.lsp4j.DidSaveTextDocumentParams;
import org.eclipse.lsp4j.DocumentFormattingParams;
import org.eclipse.lsp4j.DocumentHighlight;
import org.eclipse.lsp4j.DocumentOnTypeFormattingParams;
import org.eclipse.lsp4j.DocumentRangeFormattingParams;
import org.eclipse.lsp4j.DocumentSymbolParams;
import org.eclipse.lsp4j.Hover;
import org.eclipse.lsp4j.Location;
import org.eclipse.lsp4j.MessageParams;
import org.eclipse.lsp4j.MessageType;
import org.eclipse.lsp4j.Position;
import org.eclipse.lsp4j.PublishDiagnosticsParams;
import org.eclipse.lsp4j.Range;
import org.eclipse.lsp4j.ReferenceParams;
import org.eclipse.lsp4j.RenameParams;
import org.eclipse.lsp4j.SignatureHelp;
import org.eclipse.lsp4j.SymbolInformation;
import org.eclipse.lsp4j.TextDocumentPositionParams;
import org.eclipse.lsp4j.TextEdit;
import org.eclipse.lsp4j.WorkspaceEdit;
import org.eclipse.lsp4j.jsonrpc.messages.Either;
import org.eclipse.lsp4j.services.TextDocumentService;

public class ToulouseTextDocumentService implements TextDocumentService {

	private Map<String, String> docs = new HashMap<>();
	private ToulouseLanguageServer toulouseLanguageServer;

	public ToulouseTextDocumentService(ToulouseLanguageServer toulouseLanguageServer) {
		this.toulouseLanguageServer = toulouseLanguageServer;
	}

	private String findWord(TextDocumentPositionParams params) throws IOException {
		Position position = params.getPosition();
		String content = docs.get(params.getTextDocument().getUri());
		int startOffset = toOffset(position, content);
		int endOffset = startOffset;
		while (startOffset > 0 && Character.isJavaIdentifierPart(content.charAt(startOffset - 1))) {
			startOffset--;
		}
		while (endOffset < content.length() && Character.isJavaIdentifierPart(content.charAt(endOffset))) {
			endOffset++;
		}
		return content.substring(startOffset, endOffset);
	}

	private int toOffset(Position position, String content) throws IOException {
		int offset = 0;
		BufferedReader reader = new BufferedReader(new StringReader(content));
		for (int lineNumber = 0; lineNumber < position.getLine(); lineNumber++) {
			String line = reader.readLine();
			offset += line.length();
			offset++; //line termination chars
		}
		return offset + position.getCharacter();
	}
	
	@Override
	public CompletableFuture<Either<List<CompletionItem>, CompletionList>> completion(
			TextDocumentPositionParams position) {
		return CompletableFuture.supplyAsync(() -> Either.forLeft(LeDictionnaireToulousain.INSTANCE.allWords.stream()
				.map(word -> {
					CompletionItem item = new CompletionItem();
					item.setLabel(word);
					item.setInsertText(word);
					return item;
				}).collect(Collectors.toList())));
	}

	@Override
	public CompletableFuture<CompletionItem> resolveCompletionItem(CompletionItem unresolved) {
		return null;
	}

	@Override
	public CompletableFuture<Hover> hover(TextDocumentPositionParams position) {
		return CompletableFuture.supplyAsync(() -> {
			Hover res = new Hover();
			try {
				res.setContents(Collections.singletonList(Either.forLeft(findWord(position))));
			} catch (IOException e) {
				e.printStackTrace();
			}
			// TODO improve me here
			return res;
		});
	}

	@Override
	public CompletableFuture<SignatureHelp> signatureHelp(TextDocumentPositionParams position) {
		return null;
	}

	@Override
	public CompletableFuture<List<? extends Location>> definition(TextDocumentPositionParams position) {
		return CompletableFuture.supplyAsync(() -> {
			try {
				String word = findWord(position);
				String content = this.docs.get(position.getTextDocument().getUri());
				BufferedReader reader = new BufferedReader(new StringReader(content));
				boolean found = false;
				int lineNumber = -1;
				String line = null;
				while (!found && (line = reader.readLine()) != null) {
					found |= line.startsWith(word + "=");
					lineNumber++;
				}
				if (found) {
					Range range = new Range(new Position(lineNumber, 0), new Position(lineNumber, word.length()));
					Location res = new Location(position.getTextDocument().getUri(), range);
					return Collections.singletonList(res);
				}
			} catch (IOException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
			return null;
		});
	}

	@Override
	public CompletableFuture<List<? extends Location>> references(ReferenceParams params) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public CompletableFuture<List<? extends DocumentHighlight>> documentHighlight(TextDocumentPositionParams position) {
		return null;
	}

	@Override
	public CompletableFuture<List<? extends SymbolInformation>> documentSymbol(DocumentSymbolParams params) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public CompletableFuture<List<? extends Command>> codeAction(CodeActionParams params) {
		return CompletableFuture.supplyAsync(() ->
			params.getContext().getDiagnostics().stream()
			.map(diagnostic -> {
				String content = this.docs.get(params.getTextDocument().getUri());
				String word = null;
				try {
					word = content.substring(
							toOffset(diagnostic.getRange().getStart(), content),
							toOffset(diagnostic.getRange().getEnd(), content));
				} catch (IOException e) {
					e.printStackTrace();
				}
				return LeDictionnaireToulousain.INSTANCE.findReplacements(word).stream().map(proposal ->
					new Command("Remplacer par " + proposal, "edit", Collections.singletonList(new TextEdit(diagnostic.getRange(), proposal)))
				);
			})
			.flatMap(it -> it)
			.collect(Collectors.toList())
		);
	}

	@Override
	public CompletableFuture<List<? extends CodeLens>> codeLens(CodeLensParams params) {
		return null;
	}

	@Override
	public CompletableFuture<CodeLens> resolveCodeLens(CodeLens unresolved) {
		return null;
	}

	@Override
	public CompletableFuture<List<? extends TextEdit>> formatting(DocumentFormattingParams params) {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public CompletableFuture<List<? extends TextEdit>> rangeFormatting(DocumentRangeFormattingParams params) {
		return null;
	}

	@Override
	public CompletableFuture<List<? extends TextEdit>> onTypeFormatting(DocumentOnTypeFormattingParams params) {
		return null;
	}

	@Override
	public CompletableFuture<WorkspaceEdit> rename(RenameParams params) {
		return null;
	}

	@Override
	public void didOpen(DidOpenTextDocumentParams params) {
		this.docs.put(params.getTextDocument().getUri(), params.getTextDocument().getText());
		CompletableFuture.runAsync(() -> validate(params.getTextDocument().getUri()));
	}

	@Override
	public void didChange(DidChangeTextDocumentParams params) {
		this.docs.put(params.getTextDocument().getUri(), params.getContentChanges().get(0).getText());
		// send notification
		CompletableFuture.runAsync(() -> validate(params.getTextDocument().getUri()));
		this.toulouseLanguageServer.client.logMessage(new MessageParams(MessageType.Error, "kikoo"));
	}

	@Override
	public void didClose(DidCloseTextDocumentParams params) {
		this.docs.remove(params.getTextDocument().getUri());
	}

	@Override
	public void didSave(DidSaveTextDocumentParams params) {
	}

	private void validate(String docUri) {
		List<Diagnostic> diagnostics = new ArrayList<>();
		String content = this.docs.get(docUri);
		for (String errorWord : LeDictionnaireToulousain.INSTANCE.errorWords.keySet()) {
			int index = -1;
			while ((index = content.indexOf(errorWord, index + 1)) != -1) {
				Diagnostic diagnostic = new Diagnostic();
				diagnostic.setMessage("Oh con! On ne dit pas `" + errorWord + "` ici!");
				diagnostic.setSeverity(DiagnosticSeverity.Error);
				diagnostic.setRange(new Range()); //TODO
				diagnostics.add(diagnostic);
			}
		}
		for (String errorWord : LeDictionnaireToulousain.INSTANCE.warningWords.keySet()) {
			int index = -1;
			while ((index = content.indexOf(errorWord, index + 1)) != -1) {
				Diagnostic diagnostic = new Diagnostic();
				diagnostic.setMessage("On n'aime pas trop `" + errorWord + "` ici.");
				diagnostic.setSeverity(DiagnosticSeverity.Warning);
				diagnostic.setRange(new Range()); //TODO
				diagnostics.add(diagnostic);
			}
		}
		this.toulouseLanguageServer.client.publishDiagnostics(new PublishDiagnosticsParams(docUri, diagnostics));
	}

	
}
